@@ -1,6 +1,4 @@
 /*
-
-
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
@@ -18,22 +16,17 @@ package controllers
 
 import (
 	"context"
-	"encoding/json"
 	"fmt"
-	"reflect"
-	"sort"
-	"strings"
 	"time"
 
 	"github.com/argoproj-labs/applicationset/pkg/generators"
 	"github.com/argoproj-labs/applicationset/pkg/utils"
-	"github.com/argoproj/argo-cd/common"
-	argov1alpha1 "github.com/argoproj/argo-cd/pkg/apis/application/v1alpha1"
-	"github.com/argoproj/argo-cd/util/argo"
-	"github.com/argoproj/argo-cd/util/db"
+	argov1alpha1 "github.com/argoproj/argo-cd/v2/pkg/apis/application/v1alpha1"
+	"github.com/argoproj/argo-cd/v2/util/db"
 	"github.com/go-logr/logr"
 	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/client-go/kubernetes"
 	"k8s.io/client-go/tools/record"
 	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
 	"sigs.k8s.io/controller-runtime/pkg/handler"
@@ -45,12 +38,18 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/client"
 
 	argoprojiov1alpha1 "github.com/argoproj-labs/applicationset/api/v1alpha1"
-	appclientset "github.com/argoproj/argo-cd/pkg/client/clientset/versioned"
-	argoutil "github.com/argoproj/argo-cd/util/argo"
+	appclientset "github.com/argoproj/argo-cd/v2/pkg/client/clientset/versioned"
+	argoutil "github.com/argoproj/argo-cd/v2/util/argo"
 
 	apierr "k8s.io/apimachinery/pkg/api/errors"
+)
 
-	"github.com/imdario/mergo"
+const (
+	// Rather than importing the whole argocd-notifications controller, just copying the const here
+	//   https://github.com/argoproj-labs/argocd-notifications/blob/33d345fa838829bb50fca5c08523aba380d2c12b/pkg/controller/subscriptions.go#L12
+	//   https://github.com/argoproj-labs/argocd-notifications/blob/33d345fa838829bb50fca5c08523aba380d2c12b/pkg/controller/state.go#L17
+	NotifiedAnnotationKey             = "notified.notifications.argoproj.io"
+	ReconcileRequeueOnValidationError = time.Minute * 3
 )
 
 // ApplicationSetReconciler reconciles a ApplicationSet object
@@ -62,6 +61,7 @@ type ApplicationSetReconciler struct {
 	Generators       map[string]generators.Generator
 	ArgoDB           db.ArgoDB
 	ArgoAppClientset appclientset.Interface
+	KubeClientset    kubernetes.Interface
 	utils.Policy
 	utils.Renderer
 }
@@ -87,7 +87,7 @@ func (r *ApplicationSetReconciler) Reconcile(ctx context.Context, req ctrl.Reque
 	}
 
 	// Log a warning if there are unrecognized generators
-	checkInvalidGenerators(&applicationSetInfo)
+	utils.CheckInvalidGenerators(&applicationSetInfo)
 
 	// desiredApplications is the main list of all expected Applications from all generators in this appset.
 	desiredApplications, err := r.generateApplications(applicationSetInfo)
@@ -96,19 +96,15 @@ func (r *ApplicationSetReconciler) Reconcile(ctx context.Context, req ctrl.Reque
 	}
 
 	if validateError := r.validateGeneratedApplications(ctx, desiredApplications, applicationSetInfo, req.Namespace); validateError != nil {
-		// The reconciler presumes that any errors that are returned are a signal
-		// that the resource should attempt to be reconciled again (causing
-		// Reconcile to be called again, which will return the same error, ad
-		// infinitum until we are exponentially backed off).
+		// While some generators may return an error that requires user intervention,
+		// other generators reference external resources that may change to cause
+		// the error to no longer occur. We thus log the error and requeue
+		// with a timeout to give this another shot at a later time.
 		//
-		// In this case, since we know that what the user provided is incorrect
-		// (we successfully generated and templated their ApplicationSet, but the
-		// result of that was bad), no matter how many times we try to do so it
-		// will fail. So just log it and return that the resource was
-		// successfully reconciled (which is true... it was reconciled to an
-		// error condition).
-		log.Errorf("%s", validateError.Error())
-		return ctrl.Result{}, nil
+		// Changes to watched resources will cause this to be reconciled sooner than
+		// the RequeueAfter time.
+		log.Errorf("error occurred during application generation: %s", validateError.Error())
+		return ctrl.Result{RequeueAfter: ReconcileRequeueOnValidationError}, nil
 	}
 
 	if r.Policy.Update() {
@@ -156,7 +152,7 @@ func (r *ApplicationSetReconciler) validateGeneratedApplications(ctx context.Con
 			return err
 		}
 
-		if err := argoutil.ValidateDestination(ctx, &app.Spec.Destination, r.ArgoDB); err != nil {
+		if err := utils.ValidateDestination(ctx, &app.Spec.Destination, r.KubeClientset, namespace); err != nil {
 			return fmt.Errorf("application destination spec is invalid: %s", err.Error())
 		}
 
@@ -165,7 +161,7 @@ func (r *ApplicationSetReconciler) validateGeneratedApplications(ctx context.Con
 			return err
 		}
 		if len(conditions) > 0 {
-			return fmt.Errorf("application spec is invalid: %s", argo.FormatAppConditions(conditions))
+			return fmt.Errorf("application spec is invalid: %s", argoutil.FormatAppConditions(conditions))
 		}
 
 	}
@@ -173,90 +169,6 @@ func (r *ApplicationSetReconciler) validateGeneratedApplications(ctx context.Con
 	return nil
 }
 
-// Log a warning if there are unrecognized generators
-func checkInvalidGenerators(applicationSetInfo *argoprojiov1alpha1.ApplicationSet) {
-	hasInvalidGenerators, invalidGenerators := invalidGenerators(applicationSetInfo)
-	if len(invalidGenerators) > 0 {
-		gnames := []string{}
-		for n := range invalidGenerators {
-			gnames = append(gnames, n)
-		}
-		sort.Strings(gnames)
-		aname := applicationSetInfo.ObjectMeta.Name
-		msg := "ApplicationSet %s contains unrecognized generators: %s"
-		log.Warnf(msg, aname, strings.Join(gnames, ", "))
-	} else if hasInvalidGenerators {
-		name := applicationSetInfo.ObjectMeta.Name
-		msg := "ApplicationSet %s contains unrecognized generators"
-		log.Warnf(msg, name)
-	}
-}
-
-// Return true if there are unknown generators specified in the application set.  If we can discover the names
-// of these generators, return the names as the keys in a map
-func invalidGenerators(applicationSetInfo *argoprojiov1alpha1.ApplicationSet) (bool, map[string]bool) {
-	names := make(map[string]bool)
-	hasInvalidGenerators := false
-	for index, generator := range applicationSetInfo.Spec.Generators {
-		v := reflect.Indirect(reflect.ValueOf(generator))
-		found := false
-		for i := 0; i < v.NumField(); i++ {
-			field := v.Field(i)
-			if !field.CanInterface() {
-				continue
-			}
-			if !reflect.ValueOf(field.Interface()).IsNil() {
-				found = true
-				break
-			}
-		}
-		if !found {
-			hasInvalidGenerators = true
-			addInvalidGeneratorNames(names, applicationSetInfo, index)
-		}
-	}
-	return hasInvalidGenerators, names
-}
-
-func addInvalidGeneratorNames(names map[string]bool, applicationSetInfo *argoprojiov1alpha1.ApplicationSet, index int) {
-	// The generator names are stored in the "kubectl.kubernetes.io/last-applied-configuration" annotation
-	config := applicationSetInfo.ObjectMeta.Annotations["kubectl.kubernetes.io/last-applied-configuration"]
-	var values map[string]interface{}
-	err := json.Unmarshal([]byte(config), &values)
-	if err != nil {
-		log.Warnf("couldn't unmarshal kubectl.kubernetes.io/last-applied-configuration: %+v", config)
-		return
-	}
-
-	spec, ok := values["spec"].(map[string]interface{})
-	if !ok {
-		log.Warn("coundn't get spec from kubectl.kubernetes.io/last-applied-configuration annotation")
-		return
-	}
-
-	generators, ok := spec["generators"].([]interface{})
-	if !ok {
-		log.Warn("coundn't get generators from kubectl.kubernetes.io/last-applied-configuration annotation")
-		return
-	}
-
-	if index >= len(generators) {
-		log.Warnf("index %d out of range %d for generator in kubectl.kubernetes.io/last-applied-configuration", index, len(generators))
-		return
-	}
-
-	generator, ok := generators[index].(map[string]interface{})
-	if !ok {
-		log.Warn("coundn't get generator from kubectl.kubernetes.io/last-applied-configuration annotation")
-		return
-	}
-
-	for key := range generator {
-		names[key] = true
-		break
-	}
-}
-
 func hasDuplicateNames(applications []argov1alpha1.Application) (bool, string) {
 	nameSet := map[string]struct{}{}
 	for _, app := range applications {
@@ -268,31 +180,13 @@ func hasDuplicateNames(applications []argov1alpha1.Application) (bool, string) {
 	return false, ""
 }
 
-func (r *ApplicationSetReconciler) GetRelevantGenerators(requestedGenerator *argoprojiov1alpha1.ApplicationSetGenerator) []generators.Generator {
-	var res []generators.Generator
-
-	v := reflect.Indirect(reflect.ValueOf(requestedGenerator))
-	for i := 0; i < v.NumField(); i++ {
-		field := v.Field(i)
-		if !field.CanInterface() {
-			continue
-		}
-
-		if !reflect.ValueOf(field.Interface()).IsNil() {
-			res = append(res, r.Generators[v.Type().Field(i).Name])
-		}
-	}
-
-	return res
-}
-
 func (r *ApplicationSetReconciler) getMinRequeueAfter(applicationSetInfo *argoprojiov1alpha1.ApplicationSet) time.Duration {
 	var res time.Duration
 	for _, requestedGenerator := range applicationSetInfo.Spec.Generators {
 
-		generators := r.GetRelevantGenerators(&requestedGenerator)
+		relevantGenerators := generators.GetRelevantGenerators(&requestedGenerator, r.Generators)
 
-		for _, g := range generators {
+		for _, g := range relevantGenerators {
 			t := g.GetRequeueAfter(&requestedGenerator)
 
 			if res == 0 {
@@ -313,55 +207,33 @@ func getTempApplication(applicationSetTemplate argoprojiov1alpha1.ApplicationSet
 	tmplApplication.Namespace = applicationSetTemplate.Namespace
 	tmplApplication.Name = applicationSetTemplate.Name
 	tmplApplication.Spec = applicationSetTemplate.Spec
+	tmplApplication.Finalizers = applicationSetTemplate.Finalizers
 
 	return &tmplApplication
 }
 
-func mergeGeneratorTemplate(g generators.Generator, requestedGenerator *argoprojiov1alpha1.ApplicationSetGenerator, applicationSetTemplate argoprojiov1alpha1.ApplicationSetTemplate) (argoprojiov1alpha1.ApplicationSetTemplate, error) {
-
-	// Make a copy of the value from `GetTemplate()` before merge, rather than copying directly into
-	// the provided parameter (which will touch the original resource object returned by client-go)
-	dest := g.GetTemplate(requestedGenerator).DeepCopy()
-
-	err := mergo.Merge(dest, applicationSetTemplate)
-
-	return *dest, err
-}
 func (r *ApplicationSetReconciler) generateApplications(applicationSetInfo argoprojiov1alpha1.ApplicationSet) ([]argov1alpha1.Application, error) {
-	res := []argov1alpha1.Application{}
+	var res []argov1alpha1.Application
 
 	var firstError error
 	for _, requestedGenerator := range applicationSetInfo.Spec.Generators {
-		generators := r.GetRelevantGenerators(&requestedGenerator)
-		for _, g := range generators {
-
-			// we call mergeGeneratorTemplate first because GenerateParams might be more costly so we want to fail fast if there is an error
-			mergedTemplate, err := mergeGeneratorTemplate(g, &requestedGenerator, applicationSetInfo.Spec.Template)
-			if err != nil {
-				log.WithError(err).WithField("generator", g).
-					Error("error generating params")
-				if firstError == nil {
-					firstError = err
-				}
-				continue
+		t, err := generators.Transform(requestedGenerator, r.Generators, applicationSetInfo.Spec.Template, &applicationSetInfo)
+		if err != nil {
+			log.WithError(err).WithField("generator", requestedGenerator).
+				Error("error generating application from params")
+			if firstError == nil {
+				firstError = err
 			}
+			continue
+		}
 
-			params, err := g.GenerateParams(&requestedGenerator)
-			if err != nil {
-				log.WithError(err).WithField("generator", g).
-					Error("error generating params")
-				if firstError == nil {
-					firstError = err
-				}
-				continue
-			}
+		for _, a := range t {
+			tmplApplication := getTempApplication(a.Template)
 
-			tmplApplication := getTempApplication(mergedTemplate)
-
-			for _, p := range params {
-				app, err := r.Renderer.RenderTemplateParams(tmplApplication, p)
+			for _, p := range a.Params {
+				app, err := r.Renderer.RenderTemplateParams(tmplApplication, applicationSetInfo.Spec.SyncPolicy, p)
 				if err != nil {
-					log.WithError(err).WithField("params", params).WithField("generator", g).
+					log.WithError(err).WithField("params", a.Params).WithField("generator", requestedGenerator).
 						Error("error generating application from params")
 					if firstError == nil {
 						firstError = err
@@ -370,12 +242,13 @@ func (r *ApplicationSetReconciler) generateApplications(applicationSetInfo argop
 				}
 				res = append(res, *app)
 			}
-
-			log.WithField("generator", g).Infof("generated %d applications", len(res))
-			log.WithField("generator", g).Debugf("apps from generator: %+v", res)
-
 		}
+
+		log.WithField("generator", requestedGenerator).Infof("generated %d applications", len(res))
+		log.WithField("generator", requestedGenerator).Debugf("apps from generator: %+v", res)
+
 	}
+
 	return res, firstError
 }
 
@@ -438,7 +311,16 @@ func (r *ApplicationSetReconciler) createOrUpdateInCluster(ctx context.Context,
 		action, err := utils.CreateOrUpdate(ctx, r.Client, found, func() error {
 			// Copy only the Application/ObjectMeta fields that are significant, from the generatedApp
 			found.Spec = generatedApp.Spec
+
+			// Preserve argo cd notifications state (https://github.com/argoproj-labs/applicationset/issues/180)
+			if state, exists := found.ObjectMeta.Annotations[NotifiedAnnotationKey]; exists {
+				if generatedApp.Annotations == nil {
+					generatedApp.Annotations = map[string]string{}
+				}
+				generatedApp.Annotations[NotifiedAnnotationKey] = state
+			}
 			found.ObjectMeta.Annotations = generatedApp.Annotations
+
 			found.ObjectMeta.Finalizers = generatedApp.Finalizers
 			found.ObjectMeta.Labels = generatedApp.Labels
 			return controllerutil.SetControllerReference(&applicationSet, found, r.Scheme)
@@ -498,10 +380,15 @@ func (r *ApplicationSetReconciler) getCurrentApplications(_ context.Context, app
 	return current.Items, nil
 }
 
-// deleteInCluster will delete application that are current in the cluster but not in appList.
+// deleteInCluster will delete Applications that are currently on the cluster, but not in appList.
 // The function must be called after all generators had been called and generated applications
 func (r *ApplicationSetReconciler) deleteInCluster(ctx context.Context, applicationSet argoprojiov1alpha1.ApplicationSet, desiredApplications []argov1alpha1.Application) error {
 
+	clusterList, err := utils.ListClusters(ctx, r.KubeClientset, applicationSet.Namespace)
+	if err != nil {
+		return err
+	}
+
 	// Save current applications to be able to delete the ones that are not in appList
 	current, err := r.getCurrentApplications(ctx, applicationSet)
 	if err != nil {
@@ -523,7 +410,7 @@ func (r *ApplicationSetReconciler) deleteInCluster(ctx context.Context, applicat
 		if !exists {
 
 			// Removes the Argo CD resources finalizer if the application contains an invalid target (eg missing cluster)
-			err := r.removeFinalizerOnInvalidDestination(ctx, applicationSet, &app, appLog)
+			err := r.removeFinalizerOnInvalidDestination(ctx, applicationSet, &app, clusterList, appLog)
 			if err != nil {
 				appLog.WithError(err).Error("failed to update Application")
 				if firstError != nil {
@@ -548,21 +435,54 @@ func (r *ApplicationSetReconciler) deleteInCluster(ctx context.Context, applicat
 }
 
 // removeFinalizerOnInvalidDestination removes the Argo CD resources finalizer if the application contains an invalid target (eg missing cluster)
-func (r *ApplicationSetReconciler) removeFinalizerOnInvalidDestination(ctx context.Context, applicationSet argoprojiov1alpha1.ApplicationSet, app *argov1alpha1.Application, appLog *log.Entry) error {
+func (r *ApplicationSetReconciler) removeFinalizerOnInvalidDestination(ctx context.Context, applicationSet argoprojiov1alpha1.ApplicationSet, app *argov1alpha1.Application, clusterList *argov1alpha1.ClusterList, appLog *log.Entry) error {
 
 	// Only check if the finalizers need to be removed IF there are finalizers to remove
 	if len(app.Finalizers) == 0 {
 		return nil
 	}
 
+	var validDestination bool
+
+	// Detect if the destination is invalid (name doesn't correspond to a matching cluster)
+	if err := utils.ValidateDestination(ctx, &app.Spec.Destination, r.KubeClientset, applicationSet.Namespace); err != nil {
+		appLog.Warnf("The destination cluster for %s couldn't be found: %v", app.Name, err)
+		validDestination = false
+	} else {
+
+		// Detect if the destination's server field does not match an existing cluster
+
+		matchingCluster := false
+		for _, cluster := range clusterList.Items {
+
+			// Server fields must match. Note that ValidateDestination ensures that the server field is set, if applicable.
+			if app.Spec.Destination.Server != cluster.Server {
+				continue
+			}
+
+			// The name must match, if it is not empty
+			if app.Spec.Destination.Name != "" && cluster.Name != app.Spec.Destination.Name {
+				continue
+			}
+
+			matchingCluster = true
+			break
+		}
+
+		if !matchingCluster {
+			appLog.Warnf("A match for the destination cluster for %s, by server url, couldn't be found.", app.Name)
+		}
+
+		validDestination = matchingCluster
+	}
 	// If the destination is invalid (for example the cluster is no longer defined), then remove
 	// the application finalizers to avoid triggering Argo CD bug #5817
-	if err := argoutil.ValidateDestination(ctx, &app.Spec.Destination, r.ArgoDB); err != nil {
+	if !validDestination {
 
 		// Filter out the Argo CD finalizer from the finalizer list
-		newFinalizers := []string{}
+		var newFinalizers []string
 		for _, existingFinalizer := range app.Finalizers {
-			if existingFinalizer != common.ResourcesFinalizerName { // only remove this one
+			if existingFinalizer != argov1alpha1.ResourcesFinalizerName { // only remove this one
 				newFinalizers = append(newFinalizers, existingFinalizer)
 			}
 		}
